---
title: Web APIs
type: api
layout: docs
order: 1
parent_section: api
---

## WebVR

WebVR is an open specification that makes it possible to experience VR in your browser. The goal is to make it easier for everyone to get into VR experiences, no matter what device you have.

Read more at the [WebVR](https://developer.mozilla.org/en-US/docs/Web/API/WebVR_API) API documentation.

## WebXR

WebXR is driving support for accessing virtual reality (VR) and augmented reality (AR) devices, including sensors and head-mounted displays on the Web.

Read more at the [WebXR](https://www.w3.org/TR/webxr/) API documentation.

## React

React is a JavaScript library for building user interfaces. Learn what React is all about on their homepage.

Read more at the [React](https://reactjs.org/docs/react-api.html) API documentation.

## HTML5

HTML5 is the latest evolution of the standard that defines HTML. The term represents two different concepts. It is a new version of the language HTML, with new elements, attributes, and behaviors, and a larger set of technologies that allows the building of more diverse and powerful Web sites and applications. This set is sometimes called HTML5 & friends and often shortened to just HTML5.

Read more at the [HTML5](https://developer.mozilla.org/en-US/docs/Web/Guide/HTML/HTML5) API documentation.

## CSS
Cascading Style Sheets (CSS) is a stylesheet language used to describe the presentation of a document written in HTML or XML (including XML dialects such as SVG or XHTML). CSS describes how elements should be rendered on screen, on paper, in speech, or on other media.

Read more at the [CSS](https://developer.mozilla.org/en-US/docs/Web/API/CSS) API documentation.

## THREE.js

Three.js is a cross-browser JavaScript library and Application Programming Interface (API) used to create and display animated 3D computer graphics in a web browser. Three.js uses WebGL(see WebGL in Native APIs).

Read more at the [THREE.js](https://threejs.org/docs/) API documentation.

## Express
Express is for making JS servers to assist Exokit in creating HTTP requests and other server functions.

Read more at the [Express](https://expressjs.com/en/api.html) API documentation.

## WebSockets
WebSockets are an advanced technology that makes it possible to open an interactive communication session between the user's browser and a server. With this API, you can send messages to a server and receive event-driven responses without having to poll the server for a reply.

Read more at the [WebSockets](https://developer.mozilla.org/en-US/docs/Web/API/WebSockets_API) API documentation.


## Fetch API

The Fetch API provides an interface for fetching resources (including across the network). It will seem familiar to anyone who has used XMLHttpRequest, but the new API provides a more powerful and flexible feature set.

Read more at the [Fetch API](https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API) API documentation.

## Web Workers

Web Workers makes it possible to run a script operation in a background thread separate from the main execution thread of a web application. The advantage of this is that laborious processing can be performed in a separate thread, allowing the main (usually the UI) thread to run without being blocked/slowed down.

Read more at the [Web Workers](https://developer.mozilla.org/en-US/docs/Web/API/Web_Workers_API/Using_web_workers) API documentation.

## WebGL

WebGL (Web Graphics Library) is a JavaScript API for rendering interactive 3D and 2D graphics within any compatible web browser without the use of plug-ins. WebGL does so by introducing an API that closely conforms to OpenGL ES 2.0 that can be used in HTML5 <canvas\> elements.

Read more at the [WebGL](https://developer.mozilla.org/en-US/docs/Web/API/WebGL_API) API documentation.

## Canvas

Added in HTML5, the HTML <canvas\> element can be used to draw graphics via scripting in JavaScript. For example, it can be used to draw graphs, make photo compositions, create animations, or even do real-time video processing or rendering.

Read more at the [Canvas](https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API) API documentation.

## WebAudio

The Web Audio API provides a powerful and versatile system for controlling audio on the Web, allowing developers to choose audio sources, add effects to audio, create audio visualizations, apply spatial effects (such as panning) and much more.

Read more at the [WebAudio](https://developer.mozilla.org/en-US/docs/Web/API/Web_Audio_API) API documentation.

## WebRTC

WebRTC (Web Real-Time Communications) is a technology which enables Web applications and sites to capture and optionally stream audio and/or video media, as well as to exchange arbitrary data between browsers without requiring an intermediary. The set of standards that comprises WebRTC makes it possible to share data and perform teleconferencing peer-to-peer, without requiring that the user install plug-ins or any other third-party software.

Read more at the [WebRTC](https://developer.mozilla.org/en-US/docs/Web/API/WebRTC_API) API documentation.

## More Info

For more information, check out the [API overview](../overview).
